#include<stdio.h>
#include<strings.h>
#include"floppy.h"
#include"fz1.h"

typedef struct fdata {
	short ftype;
	unsigned int ss;
	unsigned int bnum;
	unsigned int vnum;
	unsigned int wsize;
	unsigned int fsize;
	char name[14];
} fdata_t;

typedef struct bdata {
	unsigned int bstep;
	unsigned short vp[64];
	char name[14];
} bdata_t;

char *tname[]={
	"all",
	"bank",
	"voice",
	"effect",
	0,
	"external"
};

fdata_t *files;
bdata_t *banks;
unsigned char **voices, *dname, *tbuf, *map;
int occ,fcnt,i,j,k,ss,es,verbose,fsize;

int main(int argc, char **argv) {
	argc--; argv++;
	
	if (argc==1) {
		if (strcmp(*argv,"-v")==0) 
			verbose=1;
		else {
			printf("Usage: fzdir [-v]\n");
			exit(-1);
		}
	} else 
		verbose=0;

	dname=(char *)malloc(14);
	memset(dname,0,14);

	/* read disk name and count occupied blocks */
	tbuf=(char *)malloc(1024);
	sector_read(0,0,tbuf);	
	memcpy(dname,tbuf,12);
	for(i=11;i>=0;i--) {
		if (dname[i]!=0x20) break;
		dname[i]=0;
	}
	
	occ=0;
	for(i=0x80;i<0x120;i++) {
		for(j=0;j<8;j++) 
			if ((tbuf[i])&&(1<<j)) occ++;
	}
	
	
	/* read disk directory */
	files=(fdata_t *)malloc(sizeof(fdata_t)*64);
	tbuf=(char *)realloc(tbuf,2048);
	sector_read(0,1,tbuf);
	sector_read(0,2,tbuf+1024);
	fcnt=0;
	for(i=0;i<64;i++) {
		if (tbuf[i*16]!=0) {
			files[fcnt].ss=tbuf[i*16+14]|tbuf[i*16+15]<<8;
			files[fcnt].ftype=tbuf[i*16+12];
			memset(files[fcnt].name,0,14);
			memcpy(files[fcnt].name,tbuf+i*16,12);
			fcnt++;
		}
	}
	
	/* get file parameters */
	tbuf=(char *)realloc(tbuf,1024);
	for(i=0;i<fcnt;i++) {
		sector_read(0,files[i].ss,tbuf);
		files[i].bnum=tbuf[1018]|tbuf[1019]<<8;
		files[i].vnum=tbuf[1020]|tbuf[1021]<<8;
		files[i].wsize=tbuf[1022]|tbuf[1023]<<8;
	}
	
	/* list file data */
	printf("\nContents of disk %s\n",dname);
	for(i=0;i<strlen(dname)+18;i++) printf("-");
	printf("\n");
	for(i=0;i<fcnt;i++) {
		files[i].fsize=files[i].bnum*1024+(files[i].vnum>>2)*1024+files[i].wsize;
		printf("(%2d) %s (%s data)\n",i+1,files[i].name,tname[files[i].ftype]);
		printf("\tOverall size:\t%dkB\n",files[i].bnum+(files[i].vnum>>2)+files[i].wsize);
		printf("\t# of banks:\t%d\n",files[i].bnum);
		printf("\t# of voices:\t%d\n",files[i].vnum);
		printf("\tPCM data size:\t%dkB\n",files[i].wsize);
		printf("\n");
	}

	/* list file contents on verbose mode */
	if(verbose==1) {
		for(i=0;i<fcnt;i++) {

			/* read bank&voice part of file */
			fsize=files[i].bnum+(files[i].vnum>>2);
			tbuf=(char *)realloc(tbuf,fsize*1024);
			map=(char *)malloc(1024);
			sector_read(0,files[i].ss,map);
			for(j=0;j<254;j+=4) {
				ss=(map[j]|(map[j+1]<<8))&0xffff;
				if (j==0) ss++;
				es=(map[j+2]|(map[j+3]<<8))&0xffff;
				if (ss==0) 
					break;
				if ((1+es-ss)>fsize) 
					es=ss+fsize-1;					
				m_sector_read(0,ss,es,tbuf);
				if (es==ss+fsize) 
					break;
			}

			/* get voice names */
			voices=(unsigned char **)malloc(files[i].vnum*sizeof(char *));
			for(j=0;j<files[i].vnum;j++) {
				voices[j]=(char *)malloc(14);
				strcpy(voices[j],tbuf+j*256+files[i].bnum*1024+0xb2);
			}

			/* get bank names and voice assignments */
			banks=(bdata_t *)malloc(files[i].bnum*sizeof(bdata_t));
			for(j=0;j<files[i].bnum;j++) {
				banks[j].bstep=tbuf[j*1024]|(tbuf[k*1024+1]<<8);
				for(k=0;k<banks[j].bstep;k++)
					banks[j].vp[k]=tbuf[j*1024+k*2+0x202]|(tbuf[j*1024+k*2+0x203]<<8);
				strcpy(banks[j].name,tbuf+j*1024+0x282);
				for(k=11;k>=0;k--) {
					if(banks[j].name[k]!=0x20) break;
					banks[j].name[k]=0;
				}

				printf("\n\tBank %d (%s)",j+1,banks[j].name);
				for(k=0;k<banks[j].bstep;k++) {
					if (k%4==0) printf("\n\t");
					printf("%2d. %s  ",k+1,voices[banks[j].vp[k]]);
				}
				printf("\n");
			}


			if (files[i].bnum==0) {
				printf("\n\tVoices");
				for(k=0;k<files[i].vnum;k++) {
					if (k%4==0) printf("\n\t");
					printf("%2d. %s  ",k+1,voices[k]);
				}
			}
		}
		free(map);
		free(voices);
		printf("\n");
	}

	printf("Blocks used/free:\t %d/%d (%d%% used)\n",occ,1280-occ,occ*100/1280);
	
	free(dname);
	free(tbuf);
	free(files);
	free(banks);
	free(voices);
	
	exit(0);
}