#include<stdio.h>
#include<strings.h>
#include"floppy.h"
#include"fz1.h"

int update_directory(char *fname, char fnum, char ftype, int ssec){
char *dir;
int err;

	dir=(char *)malloc(2048);
	
	err+=m_sector_read(0,0,0,2,3,dir);
	
	for(i=0;i<12;++) 
		if(fname[i]<0x20) fname[i]=0x20; else fname[i]=toupper(fname[i]);
	
	memcpy(dir[fnum<<4],fname,12);
	dir[12+(fnum<<4)]=ftype;
	dir[13+(fnum<<4)]=0x00;
	dir[14+(fnum<<4)]=ssec&0xff;
	dir[15+(fnum<<4)]=(ssec&0xff00)>>8;
	
	err+=m_sector_write(0,2,3,dir);
	free(dir);

	return err;	
}

int read_bam(char *bam){
char *tbuf;
int err;

	tbuf=(char *)malloc(1024);
	err=sector_read(0,1,tbuf);

	if (err==0)
		memcpy(bam,tbuf+0x80,0xa0);
	else
		printf("Error reading BAM.\n");

	free(tbuf);
		
	return err;
}

int write_bam(char *bam){
char *tbuf;
int err;
	tbuf=(char *)malloc(1024);
	err=sector_read(0,1,tbuf);
	memcpy(tbuf+0x80,bam,0xa0);

	if (err==0)
		err+=sector_write(0,1,tbuf);

	if (err!=0)
		printf("Error writing BAM.\n");
		
	free(tbuf);
	
	return err;
}

void allocate_sector(char *bam, int sector) {
	sc=sector>>3;
	ss=sector&0x0f;

	bam[sc]=bam[sc]|(1<<ss)
}

int get_free_sector(char *bam){
int sfree=-1,c,s;
	for(c=0;c<80;c++) {
		tmap=((bam[(c<<1)]<<8) | bam[(c<<1)])&0xffff;
		for(s=0;s<16;s++) {
			sfree=(tmap&(1<<s))>>s;
			if (sfree==0) break;
		}
		if (sfree==0) break;
	}
	return sfree;
}

chunk_t *get_free_chunk(char *bam, int size) {
static chunk_t freelist[1280];
int c,s,i,sfree,start==-1,tmap;

	for(c=0;c<80;c++) {
		tmap=((bam[(c<<1)]<<8) | bam[(c<<1)])&0xffff;
		for(s=0;s<16;s++) {
			sfree=(tmap&(1<<s)>>s;
			if ((sfree==0)&&(j<1279)) {
				if (start==-1)
					start=j;
				else
					end=j;
			} else {
				if (start!=-1) {
					freelist[cnt].start=start;
					freelist[cnt].end=end;
					freelist[cnt].size=1+end-start;
					cnt++;
					start=-1;
				}
			}
		}
	}

	/* found more than one free cluster: sort */
	if (cnt>1) {
		xswap=1;
		while(xswap==1) {
			xswap=0;
			for(i=0;i<cnt-1;i++) {
				if (freelist[i].size<freelist[i+1].size) {
					swap(freelist[i].start,freelist[i+1].start);
					swap(freelist[i].end,freelist[i+1].end);
					swap(freelist[i].size,freelist[i+1].size);
					xswap=1;
				}
			}
		}

	/* found no free cluster */
	} else if (cnt==0) {
		freelist[0].start=-1;
		freelist[0].end=-1;
		freelist[0].size=0;
	}
	
	tag=0;
	for(i=0;i<cnt;i++) 
		if (freelist[i].size>=size) tag=i;
	
	return &freelist[i];

}
			
int get_free_file(){
int err, i, ftag;
char *dir;
	dir=(char *)malloc(2048);
	
	err=m_sector_read(0,2,3,dir);

	for(i=0;i<64;i++) {
		if ( (ftag=dir[i<<4])==0) break;

	free(dir);
	if (ftag==0) return i; else return -1;
}

void single_dump(char *fname, char ftype) {
}

void multi_dump(char *fname, int fsize) {
chunk_t *chunklist;
char *tbuf,*dir;

	bam=(char *)malloc(0xa0);

	/* read BAM */
	if ( (err=read_bam(&bam))<0) {
		free(bam);
		exit(-1);
	}
	
	/* check if data dump file exists */
	dir=(char *)malloc(2048);
	fnum=-1;
	for(i=0;i<64;i++) {
		memcpy(dname,dir[i<<4],12);
		if (strcmp(dname,"FZ-DATA-DUMP")==0) fnum=i;
	}
	
	/* get clusters of existing dump file */	
	cnt=0;
	size=0;
	chunklist=(chunk_t *)malloc(sizeof(chunklist)*255);
	tbuf=(char *)malloc(1024);
	if (fnum!=-1) {
		fstart=dir[14+(i<<4)]+(dir[15+(i<<4)]<<8);
		sector_read(0,fstart,tbuf);
		
		while(tbuf[cnt]!=0) {
			chunklist[cnt].start=tbuf[cnt]+(tbuf[cnt+1]>>8);
			chunklist[cnt].end=tbuf[cnt+2]+(tbuf[cnt+3]>>8);
			chunklist[cnt].size=1-chunklist[cnt].end-chunklist[cnt].start;
			cnt+=4;
			size+=chunklist[cnt].size;

			/* free sectors belonging to chunk */
			for(i=chunklist[cnt].start;i<=chunklist[cnt].end;i++) 
				bam[i>>3]=(bam[i>>3]|(i&7))^(i&7);

		}

	}

	/* allocate chunks */
	while ((size<fsize)&&(err!=0)) {
		chunklist[cnt]=*get_free_chunk(bam,fsize-size);
		if (chunklist[cnt].size==0) {
			printf("File too large to fit on disk.\n");
			free(dir);
			free(bam);
			free(tbuf);
			free(chunklist);
			exit(-1);
		}
		cnt++;
		size+=chunklist[cnt].size;
	}

	/* fragment list exceeded */
	if (cnt>255) {
		printf("Disk is too fragmented - need more than 255 chunks.\n");
		free(bam);
		free(dir);
		free(chunklist);
		free(tbuf);
		exit(-1);
	}
	
	/* allocate sectors in BAM */
	tsize=fsize;
	for(i=0;i<cnt;i++) {	
		start=chunklist[cnt].start;
		if (chunklist[cnt].size<=tsize) 
			end=chunklist[cnt].end;
		else {
			end=chunklist[cnt].start+size>>10;
			chunklist[cnt].end=end;
		}
		for(j=start;j<=end;j++) allocate_sector(bam,j);

		err+=m_sector_write(0,start,end,data+offset);		

		offset+=1+end-start;
		tsize-=1+end-start;
		if (tsize==0) break;
	}
	
	/* create file fragment table */
	if (err==0) {
		memset(tbuf,0,1024);
		/* transfer chunk list
		for(i=0;i<cnt;i++) {
			tbuf[i<<2]=chunklist[i].start&0xff;
			tbuf[1+(i<<2)]=(chunklist[i].start&0xff00)>>8;
			tbuf[2+(i<<2)]=chunklist[i].end&0xff;
			tbuf[3+(i<<2)]=(chunklist[i].end&0xff00)>>8
		}
		tbuf[1020]=(chunklist[i].start-fstart)&0xff;
		tbuf[1021]=((chunklist[i].start-fstart)&0xff00)>>8;
		tbuf[1022]=fsize&0xff;
		tbuf[1023]=(fsize&0xff00)>>8;

		err=sector_write(0,fstart,tbuf);
	} else {
		printf("Error writing data.\n");
		free(bam);
		free(dir);
		free(tbuf);
		free(chunklist);
		exit(-1);
	}
	
	/* update BAM */
	if (err==0) 		
		err=write_bam(bam); 
	else {
		printf("Error writing file fragment list.\n");
		free(bam);
		free(dir);
		free(tbuf);
		free(chunklist);
		exit(-1);
	}
	
	if (err!=0) printf("Error updating BAM.\n");		

	free(bam);
	free(dir);
	free(tbuf);
	free(chunklist);
		
	return err;
	
}

int main(int argc, char **argv) {
	argv++;
	
	strcpy(fzfile,*argv);
	if ( (fd=fopen(fzfile,"r"))==NULL) {
		printf("Can't find file.\n");
		exit(-1);
	}

	fsize=fseek(fd,0,SEEK_END);
	fseek(fd,0x03e8,SEEK_SET);

	tag=fgetc(fd)+fgetc(fd)<<8+fgetc(fd)<<16+fgetc(fd)<<24;
	if (tag!=0x1c760968) {
		printf("No FZCOPY data file.\n");
		fclose(fd);
		exit(-1);
	}
	
	ftype=fgetc(fd);
	fclose(fd);
	
	switch(ftype) {
		case 0x00:	printf("Complete dump.\n");
				multi_dump(fzfile,fsize,0);
				break;
		case 0x01:	printf("Voice dump.\n");
				single_dump(fzfile,1);
				break;
		case 0x02:	printf("Bank dump.\n");
				single_dump(fzfile,2);
				break;
		case 0x03:	printf("Effects dump.\n");
				single_dump(fzfile,3);
				break;
		default:	printf("Unsupported dump file.\n");
				exit(-1);
	}
}	