#include<stdio.h>
#include<strings.h>
#include"floppy.h"

static char *dump[]={
        "full\0",
        "bank\0",
        "voice\0",
        "effects\0"
};
                                
int err=0, i, j, blocks, fsize, ftype, start=0, pos=0;
unsigned char c, *dfile, bnum, vnum, *buf;
unsigned int wcount;
FILE *fd;

void error(char *what) {
	printf("An error occured during %s.\n",what);
	exit(-1);
}

int main(int argc, char **argv) {
	argc--; argv++;

	if ((*argv==NULL)||(strcmp(*argv,"-h")==0)) {
		printf("\n\tfzftrans - transfers FZF dump file to FZ-1 disk.\n");
		printf("\t(c)2000 by Rainer Buchty     (rainer@buchty.net)\n\n");
		printf("\tUsage: fzftrans dumpfile\n\n");
		exit(-1);
	}
	
	if ((fd=fopen(*argv,"r"))==NULL) {
		printf("Unable to open %s.\n",*argv);
		exit(-1);
	}
	
	fseek(fd,0,SEEK_END);
	fsize=ftell(fd);
	rewind(fd);

	blocks=fsize>>10;
	if (fsize%1024!=0) blocks++;
	
	printf("FZF dump file \"%s\" has %d bytes (%d blocks)\n",*argv,fsize,blocks);	

	for(i=0;i<strlen(*argv);i++) {
		(*argv)[i]=toupper((*argv)[i]);
		if ((*argv)[i]==0x2f) pos=i;
	}
	if (strcmp((*argv)+strlen(*argv)-4,".FZF")==0) memset((*argv)+strlen(*argv)-4,0,4);

	dfile=(char *)malloc(fsize);
	memset(dfile,0,blocks*1024);

	/* transfer data */
	fread(dfile,1,fsize,fd);
	ftype=dfile[0x3e8+5];
	bnum=dfile[0x3e8+7];
	vnum=dfile[0x3e8+8];
	wcount=(dfile[0x3e8+0x0d]<<8)+dfile[0x3e8+0x0c];
	printf("\tFile type: %s dump\n",dump[ftype]);
	printf("\tNumber of banks:  %4d\n",bnum);
	printf("\tNumber of voices: %4d\n",vnum);
	printf("\tSize of PCM data: %4d\n",wcount);
	
	if (blocks>1279) {
		printf("Too many blocks... Setting to 1280.\n");
		blocks=1276;
	}
	
	start=0;
	for(i=start;i<blocks+start;i++) {
		printf("Writing sector %d...\r",i+4); fflush(stdout);
 		err+=sector_write(0,i+4,dfile+i*1024);
	}
	if (err!=0) error("data transfer");	

	/* create fragment list */
	printf("Creating fragment list...\r");
	memset(dfile,0,1024);
	dfile[0]=3; 			// start sector = 3
	dfile[2]=(3+blocks)&0xff;	// end sector=3+blocks
	dfile[3]=(3+blocks)>>8; 	//
	dfile[1018]=bnum;		// # of banks
	dfile[1020]=vnum;		// # of voices
	dfile[1022]=(wcount&0xff);	// PCM size
	dfile[1023]=(wcount>>8)&0xff;	//
	if (sector_write(0,3,dfile)!=0) error("fragment list creation");
	
	/* update directory */
	printf("Updating directory...\r");
	memset(dfile,0,1024);
	if (sector_write(0,2,dfile)!=0) error("directory write");

	memcpy(dfile,"FULL-DATA-FZ",12);
	dfile[12]=ftype;		// file type
	dfile[13]=0;			//
	dfile[14]=3;			// data fragment list at sector 3
	dfile[15]=0;			//
	if (sector_write(0,1,dfile)!=0) error("directory write");
	
	memset(dfile,0,0x120);
	memset(dfile+0x120,0xff,0x2e0);

	/* create BAM */
	for(i=0;i<blocks+3;i+=8) {
		for(j=0;j<8;j++) if((i>>3)*8+j<blocks+3) dfile[0x80+(i>>3)]=dfile[0x80+(i>>3)]|(1<<j);
	}
			
	/* update disk name & work pass */
	for(i=0;i<12;i++) {
		if (i>strlen(*argv+pos)) 
			c=0x20; 
		else if ( (c=(*argv+pos)[i])<0x20) c=0x20; 
		dfile[i]=c;
		dfile[i+0x10]=c;
	}
	dfile[0x0e]=2;

	if (sector_write(0,0,dfile)!=0) error("disk name & BAM write");

	free(dfile);
	printf("Done.                    \n");	
	exit(0);
}		